package nodeinfo

import (
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"net/http"
	"net/url"
	"strings"
)

type fetcher struct {
	client   *http.Client
	matrix   bool
	nodeinfo bool
}

type Fetcher interface {
	QueryNode(domain string) ([]*Node, error)
}

func NewFetcher(opts ...QueryNodeOption) (Fetcher, error) {
	f := &fetcher{
		// don't follow redirects
		client: &http.Client{
			CheckRedirect: func(req *http.Request, via []*http.Request) error {
				return http.ErrUseLastResponse
			},
		},
		matrix:   true,
		nodeinfo: true,
	}

	for _, fn := range opts {
		if err := fn(f); err != nil {
			return nil, fmt.Errorf("could not apply option: %w", err)
		}
	}
	return f, nil
}

func (f *fetcher) validate(domain string) (string, error) {
	// validate options
	if !(f.nodeinfo || f.matrix) {
		return "", fmt.Errorf("all supported discovery protocols are disabled")
	}

	// validate user input
	if !(strings.HasPrefix(domain, "http://") || strings.HasPrefix(domain, "https://")) {
		domain = "https://" + domain
	}
	urlDomain, err := url.Parse(domain)
	if err != nil {
		return "", fmt.Errorf("parsing domain failed: %w", err)
	}

	return urlDomain.Host, nil
}

func (f *fetcher) QueryNode(rawDomain string) ([]*Node, error) {
	domain, err := f.validate(rawDomain)
	if err != nil {
		return nil, fmt.Errorf("validation of fetcher failed: %w", err)
	}

	var el []error
	var infos []*Node

	if f.nodeinfo {
		n, found, err := f.doNodeInfo(domain)
		if found && n != nil {
			infos = append(infos, n)
		}
		el = append(el, err)

		if !found {
			n, found, err = f.doNodeInfoX(domain)
			if found && n != nil {
				infos = append(infos, n)
			}
			el = append(el, err)
		}
	}

	if f.matrix {
		n, found, err := f.doMatrix(domain)
		if found && n != nil {
			infos = append(infos, n)
		}
		el = append(el, err)
	}

	if len(infos) != 0 {
		return infos, nil
	}

	return nil, fmt.Errorf("could not detect well-known service: %w", errors.Join(el...))
}

type nodeInfoLookup struct {
	Links []struct {
		Schema string `json:"rel"`
		Href   string `json:"href"`
	} `json:"links"`
}

type nodeInfoV1 struct {
	Software          Software               `json:"software"`
	Version           string                 `json:"version"`
	OpenRegistrations bool                   `json:"openRegistrations"`
	Protocols         Services               `json:"protocols"`
	Services          Services               `json:"services"`
	Usage             Usage                  `json:"usage"`
	Metadata          map[string]interface{} `json:"metadata"`
}

func (f *fetcher) doNodeInfo(domain string) (*Node, bool, error) {
	// A client should first try the HTTPS protocol and fall back to HTTP on connection errors or if it can't validate the presented certificate.
	// TODO: check if https or http
	schema := "https"

	resp, err := f.client.Get(fmt.Sprintf("%s://%s/.well-known/nodeinfo", schema, domain))
	if err != nil {
		return nil, false, err
	}

	// no nodeinfo server found
	if resp.StatusCode == http.StatusNotFound {
		return nil, false, nil
	} else if resp.StatusCode != http.StatusOK {
		return nil, false, fmt.Errorf("unexpected status code on lookup")
	}

	// get nodeinfo endpoints
	rawResp, err := io.ReadAll(resp.Body)
	if err != nil {
		return nil, true, err
	}
	lookup := nodeInfoLookup{}
	if err := json.Unmarshal(rawResp, &lookup); err != nil {
		return nil, true, err
	}

	links := make(map[string]string)
	for _, link := range lookup.Links {
		links[link.Schema] = link.Href
	}
	if len(links) == 0 {
		return nil, true, fmt.Errorf("broken lookup endpoint, no links found")
	}

	var href string
	isV1 := false
	if ref, ok := links["http://nodeinfo.diaspora.software/ns/schema/2.1"]; ok {
		href = ref
	} else if ref, ok := links["http://nodeinfo.diaspora.software/ns/schema/2.0"]; ok {
		href = ref
	} else if ref, ok := links["http://nodeinfo.diaspora.software/ns/schema/1.1"]; ok {
		isV1 = true
		href = ref
	} else if ref, ok := links["http://nodeinfo.diaspora.software/ns/schema/1.0"]; ok {
		isV1 = true
		href = ref
	} else {
		return nil, true, fmt.Errorf("broken lookup endpoint, unknown schema(s) found: %s", links)
	}

	resp, err = f.client.Get(href)
	if err != nil {
		return nil, false, err
	}
	rawResp, err = io.ReadAll(resp.Body)
	if err != nil {
		return nil, true, err
	}

	node := &Node{}
	if !isV1 {
		if err := json.Unmarshal(rawResp, node); err != nil {
			return nil, true, err
		}
	} else {
		v1Node := &nodeInfoV1{}
		if err := json.Unmarshal(rawResp, v1Node); err != nil {
			return nil, true, err
		}

		// oh no it's socialhome
		if v1Node.Software.Name == "socialhome" {
			n, _, err := f.doNodeInfoX(domain)
			if err != nil {
				return nil, true, err
			}
			if n != nil {
				return n, true, nil
			}
		}

		protocolMap := make(map[string]struct{})
		for _, v := range append(v1Node.Protocols.Inbound, v1Node.Protocols.Outbound...) {
			protocolMap[v] = struct{}{}
		}
		var protocolList []string
		for k := range protocolMap {
			protocolList = append(protocolList, k)
		}
		node = &Node{
			Version:           v1Node.Version,
			Software:          v1Node.Software,
			OpenRegistrations: v1Node.OpenRegistrations,
			Protocols:         protocolList,
			Services:          v1Node.Services,
			Usage:             v1Node.Usage,
			Metadata:          v1Node.Metadata,
		}
	}

	return node, true, nil
}

type xNodeInfo struct {
	Server struct {
		Software string `json:"software"`
		Version  string `json:"version"`
		Name     string `json:"name"`
		BaseURL  string `json:"baseUrl"`
	} `json:"server"`
	Version           string   `json:"version"`
	OpenRegistrations bool     `json:"openRegistrations"`
	Protocols         []string `json:"protocols"`
	Services          Services `json:"services"`
	Usage             struct {
		Users struct {
			Total          int64 `json:"total"`
			ActiveHalfyear int64 `json:"activeHalfyear"`
			ActiveMonth    int64 `json:"activeMonth"`
			ActiveWeek     int64 `json:"activeWeek"`
		} `json:"users"`
	} `json:"usage"`
	Relay string `json:"relay"`
}

func (f *fetcher) doNodeInfoX(domain string) (*Node, bool, error) {
	// A client should first try the HTTPS protocol and fall back to HTTP on connection errors or if it can't validate the presented certificate.
	// TODO: check if https or http
	schema := "https"

	resp, err := f.client.Get(fmt.Sprintf("%s://%s/.well-known/x-nodeinfo2", schema, domain))
	if err != nil {
		return nil, false, err
	}

	// no x-nodeinfo server found
	if resp.StatusCode == http.StatusNotFound {
		return nil, false, nil
	} else if resp.StatusCode != http.StatusOK {
		return nil, false, fmt.Errorf("unexpected status code on lookup")
	}

	// get nodeinfo endpoints
	rawResp, err := io.ReadAll(resp.Body)
	if err != nil {
		return nil, true, err
	}
	node := &xNodeInfo{}
	if err := json.Unmarshal(rawResp, node); err != nil {
		return nil, true, err
	}

	infoNode := &Node{
		Version: "2.1",
		Software: Software{
			Name:    node.Server.Software,
			Version: node.Server.Version,
		},
		OpenRegistrations: node.OpenRegistrations,
		Protocols:         node.Protocols,
		Services:          node.Services,
		Usage: Usage{
			Users: UsersUsage{
				Total:          node.Usage.Users.Total,
				ActiveHalfyear: node.Usage.Users.ActiveHalfyear,
				ActiveMonth:    node.Usage.Users.ActiveMonth,
			},
		},
		Metadata: map[string]interface{}{
			"nodeName":             node.Server.Name,
			"baseUrl":              node.Server.BaseURL,
			"usageUsersActiveWeek": node.Usage.Users.ActiveWeek,
			"relay":                node.Relay,
		},
	}

	return infoNode, true, nil
}

type matrixLookup struct {
	ServerAddr string `json:"m.server"`
}

type matrixVersion struct {
	Server struct {
		ImplName    string `json:"name"`
		ImplVersion string `json:"version"`
	} `json:"server"`
}

func (f *fetcher) doMatrix(domain string) (*Node, bool, error) {
	schema := "https"
	node := Node{
		Version:   "2.1",
		Protocols: []string{ProtocolMatrix},
	}

	// TODO: respect DNS lookup as per https://spec.matrix.org/v1.8/server-server-api/#resolving-server-names

	resp, err := f.client.Get(fmt.Sprintf("%s://%s/.well-known/matrix/server", schema, domain))
	if err != nil {
		return nil, false, err
	}

	// no matrix server found
	if resp.StatusCode == http.StatusNotFound {
		return nil, false, nil
	} else if resp.StatusCode != http.StatusOK {
		return nil, false, fmt.Errorf("unexpected status code on lookup")
	}

	// get matrix server address
	rawResp, err := io.ReadAll(resp.Body)
	if err != nil {
		return nil, true, err
	}
	lookup := matrixLookup{}
	if err := json.Unmarshal(rawResp, &lookup); err != nil {
		return nil, true, err
	}

	// get server software
	resp, err = f.client.Get(fmt.Sprintf("%s://%s/_matrix/federation/v1/version", schema, lookup.ServerAddr))
	if err != nil {
		return nil, false, err
	}
	rawResp, err = io.ReadAll(resp.Body)
	if err != nil {
		return nil, true, err
	}
	version := matrixVersion{}
	if err := json.Unmarshal(rawResp, &version); err != nil {
		return nil, true, err
	}
	node.Software.Version = version.Server.ImplVersion
	node.Software.Name = version.Server.ImplName
	if soft, found := mapMatrixSoftware[strings.ToLower(version.Server.ImplName)]; found {
		node.Software.Homepage = soft.www
		node.Software.Repository = soft.git
	}

	// TODO: old trick (https://codeberg.org/thefederationinfo/py2-federation/src/branch/master/federation/hostmeta/parsers.py#L141-L144) don't work anymore
	node.OpenRegistrations = false

	return &node, true, nil
}
